package positron;

import static android.view.KeyEvent.KEYCODE_DPAD_CENTER;
import static java.lang.Thread.sleep;

import java.io.ByteArrayOutputStream;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import junit.framework.TestSuite;
import positron.harness.DatabaseFixture;
import positron.harness.Finish;
import positron.harness.InstrumentedTestResult;
import positron.harness.InstrumentedTestRunner;
import positron.harness.PauseButton;
import positron.harness.ViewShorthand;
import android.app.Activity;
import android.app.Instrumentation;
import android.content.Intent;
import android.graphics.Bitmap;
import android.os.Bundle;
import android.util.Log;
import android.view.KeyEvent;
import android.view.View;

/**
 * Subclass this to test your application with Positron.
 *
 * First, implement suite() and return a junit TestSuite of the tests to run.
 * Your tests will want to extend positron.TestCase.
 *
 * Declare the subclass in your AndroidManifest.xml with <instrumentation>:
 * <instrumentation class=".Positron"
 *                  android:targetPackage="your.package"
 *                  android:label="My Acceptance Tests"/>
 *
 * Notice you must put your subclass inside your.package in this example.
 *
 * @author philhsmith
 */
public abstract class Positron extends Instrumentation {
	public final static String TAG = "Positron";

	private LinkedList<Activity> activities;
	private DatabaseFixture databaseFixture;
	private PauseButton pauseButton;
	private ViewShorthand viewShorthand;

	@Override
	public void onCreate(Bundle arguments) {
		super.onCreate(arguments);
		start();
	}

	@Override
	public void onStart() {
		activities = new LinkedList<Activity>();
		databaseFixture = new DatabaseFixture(this);
		pauseButton = new PauseButton(this);
		viewShorthand = new ViewShorthand();

		backup();

        ByteArrayOutputStream testOutput = new ByteArrayOutputStream(1024);
        InstrumentedTestRunner runner = new InstrumentedTestRunner(this, new PrintStream(testOutput));
        InstrumentedTestResult result = (InstrumentedTestResult)runner.doRun(suite());

        Log.i(TAG, testOutput.toString());

        finishAll();
        pauseButton.quit();

        restore();
        
        waitForIdleSync();
        finish(result.errorCount() == 0 && result.failureCount() == 0 ? 0 : 1, result.toBundle());
	}

	@Override
	public void callActivityOnResume(Activity activity) {
		super.callActivityOnResume(activity);
		Log.v(TAG, "Noticed " + activity.getClass().getName() + " resumed.");
		synchronized (activities) {
			if (activities.contains(activity)) activities.remove(activity);
			activities.addFirst(activity);
		}
	}

	/** Get a snapshot of the known activity stack. */
	public List<Activity> activities() {
		synchronized (activities) {
			for (Iterator<Activity> i = activities.iterator(); i.hasNext(); ) {
				Activity finished = i.next();
				if (finished.isFinishing()) {
					Log.v(TAG, "Noticed " + finished.getClass().getName() + " finished.");
					i.remove();
				}
			}

			return new ArrayList<Activity>(activities);
		}
	}

	public Activity activity() {
		synchronized (activities) {
			return activities.get(0);
		}
	}

	public void finish(Activity activity) {
		boolean wasPaused = paused();
		if (wasPaused) resume();
		runOnMainSync(new Finish(activity));
		if (wasPaused) pause();
	}

	public void finish() {
		finish(activity());
	}

	public void finishAll() {
		resume();
		for (Activity activity : activities()) finish(activity);
		try {
			for (int tries = 0; tries < 20 && !activities().isEmpty(); tries++) sleep(250);
		} catch (InterruptedException e) {
			Log.e(TAG, "Interrupted while waiting for all activities to finish.", e);
		}
		if (!activities().isEmpty()) Log.e(TAG, "Timed out waiting for all activities to finish.");
	}

	public void pause() {
		waitForIdleSync();
		pauseNow();
	}

	public void pauseNow() {
		pauseButton.pause();
	}

	public void resume() {
		pauseButton.resume();
	}

	public boolean paused() {
		return pauseButton.paused();
	}

	/** Backup all databases in the tested application. */
	public void backup() {
		databaseFixture.backup();
	}

	/** Backup the given database. */
	public void backup(String database) {
		databaseFixture.backup(database);
	}

	/** Execute the given Sql scripts (as raw resources in the target context.) */
	public void execSql(String database, int... scriptsAsRawResources) {
		databaseFixture.execSql(database, scriptsAsRawResources);
	}

	/** Restore all databases in the tested application that have backups. */
	public void restore() {
		databaseFixture.restore();
	}

	/**
	 * Restore the given database from a previous backup call.
	 * This deletes the backup as well.
	 */
	public void restore(String database) {
		databaseFixture.restore(database);
	}

	/** @return A TestSuite of positron.TestCases to run. */
	protected abstract TestSuite suite();

	/**
	 * Shortcut to startActivitySync the named class in the target context.
	 * @param className fully-qualified class name of the action to start.
	 */
	protected Activity startActivitySync(String className) {
		Intent intent = new Intent(Intent.MAIN_ACTION);
        intent.addLaunchFlags(Intent.NEW_TASK_LAUNCH);
        intent.setClassName(getTargetContext().getPackageName(), className);
        return startActivitySync(intent);
	}

	/** dupWindowBitmapSync, resuming momentarily if necessary. */
	public Bitmap dupWindowBitmap(Activity activity) {
		boolean wasPaused = paused();
		if (wasPaused) resume();
		Bitmap dump = dupWindowBitmapSync(activity);
		if (wasPaused) pause();
		return dump;
	}

	/** invokeMenuActionSync, resuming momentarily if necessary. */
	public void invokeMenuAction(Activity targetActivity, int id, int flag) {
		boolean wasPaused = paused();
		if (wasPaused) resume();
		invokeMenuActionSync(targetActivity, id, flag);
		if (wasPaused) pause();
	}

	/** sendCharacterSync, resuming momentarily if necessary. */
	public void sendCharacter(int keyCode) {
		boolean wasPaused = paused();
		if (wasPaused) resume();
		sendCharacterSync(keyCode);
		if (wasPaused) pause();
	}

	/** sendKeyDownUpSync, resuming momentarily if necessary. */
	public void sendKeyDownUp(int key) {
		boolean wasPaused = paused();
		if (wasPaused) resume();
		sendKeyDownUpSync(key);
		if (wasPaused) pause();
	}

	/** sendKeyDownUp(KEYCODE_DPAD_CENTER) */
	public void click() { sendKeyDownUp(KEYCODE_DPAD_CENTER); }

	/**
	 * sendKeyDownUpSync on all keys in order, resuming momentarily if necessary.
	 *
	 * Combined with the renamed constants available in positron.TestCase, this
	 * enables things like press(UP, UP, DOWN, DOWN, LEFT, RIGHT, LEFT, RIGHT, "ba")
	 * which may otherwise be cumbersome.
	 *
	 * @param keys A mixture of ints and Strings.  ints are sent with sendKeyDownUpSync,
	 * 		       Strings are send with sendStringSync.
	 * @throws IllegalArgumentException if something other than an int or String was passed.
	 */
	public void press(Object... keys) {
		boolean wasPaused = paused();
		if (wasPaused) resume();
		for (Object key : keys) {
			if (key instanceof Integer) sendKeyDownUpSync((Integer)key);
			else if (key instanceof String) sendStringSync((String)key);
		}
		if (wasPaused) pause();
	}

	/**
	 * Evaluate the ViewShorthand path starting from the given point.
	 * @return The result as the passed type.
	 */
	public <T> T at(Class<T> asA, Object from, String path) {
		return viewShorthand.evaluate(asA, from, path);
	}

	/**
	 * Evaluate the ViewShorthand path starting from the current activity.
	 * @return The result as an Object.
	 */
	public Object at(String path) {
		return viewShorthand.evaluate(activity(), path);
	}
	
	/**
	 * Evaluate the ViewShorthand path starting from the current activity.
	 * @return The result as the passed type.
	 */
	public <T> T at(Class<T> asA, String path) {
		return at(asA, activity(), path);
	}
	
	/**
	 * Evaluate the ViewShorthand path starting from the current activity.
	 * @return The result as a View.
	 */
	public View viewAt(String path) {
		return at(View.class, activity(), path);
	}
	
	/**
	 * Evaluate the ViewShorthand path starting from the current activity.
	 * @return The result as a String.  If the result would be a CharSequence, call toString on it.
	 */
	public String stringAt(String path) {
		return at(CharSequence.class, activity(), path).toString();
	}
	
	/**
	 * Evaluate the ViewShorthand path starting from the current activity.
	 * @return The result as an int.
	 */
	public int intAt(String path) {
		return at(Integer.class, activity(), path);
	}
	
	/**
	 * Evaluate the ViewShorthand path starting from the current activity.
	 * @return The result as a boolean.
	 */
	public boolean booleanAt(String path) {
		return at(Boolean.class, activity(), path);
	}
	
	/**
	 * Evaluate the ViewShorthand path starting from the current activity.
	 * @return The result as a double.
	 */
	public double doubleAt(String path) {
		return at(Double.class, activity(), path);
	}
	
	/** sendKeySync, resuming momentarily if necessary. */
	public void sendKey(KeyEvent event) {
		boolean wasPaused = paused();
		if (wasPaused) resume();
		sendKeySync(event);
		if (wasPaused) pause();
	}

	/** sendStringSync, resuming momentarily if necessary. */
	public void sendString(String text) {
		boolean wasPaused = paused();
		if (wasPaused) resume();
		sendStringSync(text);
		if (wasPaused) pause();
	}

	/** startActivitySync, resuming momentarily if necessary. */
	public Activity startActivity(Intent intent) {
		boolean wasPaused = paused();
		if (wasPaused) resume();
		Activity activity = startActivitySync(intent);
		if (wasPaused) pause();
		return activity;
	}

	/** waitForIdleSync, resuming momentarily if necessary. */
	public void waitForIdle() {
		boolean wasPaused = paused();
		if (wasPaused) resume();
		waitForIdleSync();
		if (wasPaused) pause();
	}

	@Override
	public Bitmap dupWindowBitmapSync(Activity activity) {
		if (paused()) {
			String message = "The tested application is paused! Call dupWindowBitmap instead.";
			Log.e(TAG, message);
			throw new IllegalStateException(message);
		}

		try {
			return super.dupWindowBitmapSync(activity);
		} catch(RuntimeException e) {
			Log.e(TAG, e.getMessage(), e);
			throw e;
		}
	}

	@Override
	public void invokeMenuActionSync(Activity targetActivity, int id, int flag) {
		if (paused()) {
			String message = "The tested application is paused! Call invokeMenuAction instead.";
			Log.e(TAG, message);
			throw new IllegalStateException(message);
		}

		try {
			super.invokeMenuActionSync(targetActivity, id, flag);
		} catch(RuntimeException e) {
			Log.e(TAG, e.getMessage(), e);
			throw e;
		}
	}

	@Override
	public void sendCharacterSync(int keyCode) {
		if (paused()) {
			String message = "The tested application is paused! Call sendCharacter instead.";
			Log.e(TAG, message);
			throw new IllegalStateException(message);
		}

		try {
			super.sendCharacterSync(keyCode);
		} catch(RuntimeException e) {
			Log.e(TAG, e.getMessage(), e);
			throw e;
		}
	}

	@Override
	public void sendKeyDownUpSync(int key) {
		if (paused()) {
			String message = "The tested application is paused! Call sendKeyDownUp instead.";
			Log.e(TAG, message);
			throw new IllegalStateException(message);
		}

		try {
			super.sendKeyDownUpSync(key);
		} catch(RuntimeException e) {
			Log.e(TAG, e.getMessage(), e);
			throw e;
		}
	}

	@Override
	public void sendKeySync(KeyEvent event) {
		if (paused()) {
			String message = "The tested application is paused! Call sendKey instead.";
			Log.e(TAG, message);
			throw new IllegalStateException(message);
		}

		try {
			super.sendKeySync(event);
		} catch(RuntimeException e) {
			Log.e(TAG, e.getMessage(), e);
			throw e;
		}
	}

	@Override
	public void sendStringSync(String text) {
		if (paused()) {
			String message = "The tested application is paused! Call sendString instead.";
			Log.e(TAG, message);
			throw new IllegalStateException(message);
		}

		try {
			super.sendStringSync(text);
		} catch(RuntimeException e) {
			Log.e(TAG, e.getMessage(), e);
			throw e;
		}
	}

	@Override
	public Activity startActivitySync(Intent intent) {
		if (paused()) {
			String message = "The tested application is paused! Call startActivity instead.";
			Log.e(TAG, message);
			throw new IllegalStateException(message);
		}

		try {
			return super.startActivitySync(intent);
		} catch(RuntimeException e) {
			Log.e(TAG, e.getMessage(), e);
			throw e;
		}
	}

	@Override
	public void waitForIdleSync() {
		if (paused()) {
			String message = "The tested application is paused! Call waitForIdle instead.";
			Log.e(TAG, message);
			throw new IllegalStateException(message);
		}

		try {
			super.waitForIdleSync();
		} catch(RuntimeException e) {
			Log.e(TAG, e.getMessage(), e);
			throw e;
		}
	}
}
